/**
 *  Copyright (c) 2020, Texas Instruments Incorporated
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:\
 *
 *  *   Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *  notice, this list of conditions and the following disclaimer in the
 *  documentation and/or other materials provided with the distribution.
 *  *   Neither the name of Texas Instruments Incorporated nor the names of
 *  its contributors may be used to endorse or promote products derived
 *  from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 *  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* eslint-disable @typescript-eslint/no-explicit-any */
/* eslint-disable no-empty */

import { IStorages, fs_error_codes, FS_ERROR } from './slconsts';
// eslint-disable-next-line no-duplicate-imports
import * as slconsts from './slconsts';
import { NamedDataRecord, NamedRecordFieldDescriptor } from './NamedDataRecord';

class Logger {
    info(message: string) { };
    debug(message: string) { };
    error(message: string) { };
};
const logger = new Logger();

export function setLogger(name: keyof Logger, method: (message: string) => void) {
    logger[name] = method;
};

const COMMAND_RET_API_BLOCKED = 0x50;

class FileProperties {
    constructor(public filename: string,
        public size_in_bytes = 0,
        public mirror_enable = false,
        public secure = false,
        public vendor = false,
        public is_static = false,
        public no_signature = true,
        public public_write = false,
        public public_read = false,
        public certificate_filename = '',
        public signature = '') {
    };

    token?: number;
    read_only_token?: number[];
    write_only_token?: number[];
};

function status_func(description: string, finishedCount: number, totalCount: number) {
};

export class BootLoaderNakError extends Error {
    constructor(public code: number, ...params: any) {
        super(...params);
    }
};

export class BootLoaderStatusError extends Error {
    constructor(public code: number, ...params: any) {
        super(...params);
    }
};

export class BootLoaderFileOperationError extends Error {
    constructor(public info: any, ...params: any) {
        super(...params);
    }
};

interface IVersionInfo {
    BootLoaderVersion: number;
    NwpVersion: number;
    MacVersion: number;
    PhyVersion: number;
    ChipVersion: number;
};

class VersionInfo extends NamedDataRecord<IVersionInfo> {
    littleEndian = true;
    static fieldDescriptors: NamedRecordFieldDescriptor<IVersionInfo>[] = [
        ['BootLoaderVersion', 4],
        ['NwpVersion', 4],
        ['MacVersion', 4],
        ['PhyVersion', 4],
        ['ChipVersion', 4],
        ['reserved', 4],
        ['reserved', 4]
    ];
};

interface IFileInfo {
    flags: number;
    allocated: number;
    size: number;
    max_size: number;
    master_token: number;
    write_read_token: number;
    write_token: number;
    read_token: number;
    blocks_on_device: number;
    write_counter: number;
};

class FileInfo extends NamedDataRecord<IFileInfo> {
    static fieldDescriptors: NamedRecordFieldDescriptor<IFileInfo>[] = [
        ['flags', 2],
        ['allocated', 1],
        ['padding', 1],
        ['size', 4],
        ['max_size', 4],
        ['master_token', 4],
        ['write_read_token', 4],
        ['write_token', 4],
        ['read_token', 4],
        ['blocks_on_device', 4],
        ['write_counter', 4]
    ];
};

interface IDeviceId {
    manufacturer_code: number;
    memory_type: number;
    capacity_code: number;
};

class DeviceId extends NamedDataRecord<IDeviceId> {
    static fieldDescriptors: NamedRecordFieldDescriptor<IDeviceId>[] = [
        ['reserved', 1],
        ['manufacturer_code', 1],
        ['memory_type', 1],
        ['capacity_code', 1]
    ];
};

const sflash_manufacturers_codes = new Map<number, string>([
    [0x01, 'Spansion'],
    [0x20, 'Micron'],
    [0x9D, 'ISSI'],
    [0xC2, 'Macronix'],
    [0xEF, 'Winbond']
]);

const sflash_capacity_codes = new Map<number, number>([
    [0x14, 8],
    [0x15, 16],
    [0x16, 32],
    [0x17, 64],
    [0x18, 128]
]);

const CHIP_CC3100R_FUNC_MODE = 0x00;
const CHIP_CC3100F_FUNC_MODE = 0x01;
const CHIP_CC3100Z_FUNC_MODE = 0x03;
const CHIP_CC3100RS_FUNC_MODE = 0x08;
const CHIP_CC3100FS_FUNC_MODE = 0x09;
const CHIP_CC3100ZS_FUNC_MODE = 0x0B;
const CHIP_CC3200R_FUNC_MODE = 0x10;
const CHIP_CC3200F_FUNC_MODE = 0x11;
const CHIP_CC3200Z_FUNC_MODE = 0x13;
const CHIP_CC3200RS_FUNC_MODE = 0x18;
const CHIP_CC3200FS_FUNC_MODE = 0x19;
const CHIP_CC3200ZS_FUNC_MODE = 0x1B;

enum DeviceParams {
    DEVICE_PARAM_DEFAULT_MAC_ADDR = 0x0,
    DEVICE_PARAM_PROGRAMABLE_MAC_ADDR = 0x1
};

interface IDeviceUsageInfo {
    DeviceBlockSize: number;
    DeviceBlocksCapacity: number;
    NumOfAllocatedBlocks: number;
    NumOfReservedBlocks: number;
    NumOfReservedBlocksForSystemfiles: number;
    LargestAllocatedGapInBlocks: number;
    NumOfAvailableBlocksForUserFiles: number;
};

class DeviceUsageInfo extends NamedDataRecord<IDeviceUsageInfo> {
    static fieldDescriptors: NamedRecordFieldDescriptor<IDeviceUsageInfo>[] = [
        ['DeviceBlockSize', 2],
        ['DeviceBlocksCapacity', 2],
        ['NumOfAllocatedBlocks', 2],
        ['NumOfReservedBlocks', 2],
        ['NumOfReservedBlocksForSystemfiles', 2],
        ['LargestAllocatedGapInBlocks', 2],
        ['NumOfAvailableBlocksForUserFiles', 2],
        ['padding', 2]
    ];
};

interface FileSystemInfo {
    MaxFsFiles: number;
    IsDevlopmentFormatType: number;
    Bundlestate: number;
    CC3200_BundleState: number;
    MaxFsFilesReservedForSysFiles: number;
    ActualNumOfUserFiles: number;
    ActualNumOfSysFiles: number;
    NumOfAlerts: number;
    NumOfAlertsThreshold: number;
    FATWriteCounter: number;
};

class FileSystemInfo extends NamedDataRecord<FileSystemInfo> {
    static fieldDescriptors: NamedRecordFieldDescriptor<FileSystemInfo>[] = [
        ['MaxFsFiles', 1],
        ['IsDevlopmentFormatType', 1],
        ['Bundlestate', 1],
        ['CC3200_BundleState', 1],
        ['MaxFsFilesReservedForSysFiles', 1],
        ['ActualNumOfUserFiles', 1],
        ['ActualNumOfSysFiles', 1],
        ['padding', 1],
        ['NumOfAlerts', 4],
        ['NumOfAlertsThreshold', 4],
        ['FATWriteCounter', 2],
        ['padding', 2]
    ];
};

type IFileSysControlDeviceInfo = IDeviceUsageInfo & FileSystemInfo;

class FileSysControlDeviceInfo extends NamedDataRecord<IFileSysControlDeviceInfo> {
    static fieldDescriptors: NamedRecordFieldDescriptor<IFileSysControlDeviceInfo>[] = (DeviceUsageInfo.fieldDescriptors as NamedRecordFieldDescriptor<IFileSysControlDeviceInfo>[]).concat(FileSystemInfo.fieldDescriptors);
};

interface IHeaderInfo {
    data_length: number;
    checksum: number;
};

class HeaderInfo extends NamedDataRecord<IHeaderInfo> {
    static fieldDescriptors: NamedRecordFieldDescriptor<IHeaderInfo>[] = [
        ['data_length', 2],
        ['checksum', 1]
    ];
};

interface IStorageInfo {
    block_size: number;
    n_blocks: number;
};

class StorageInfo extends NamedDataRecord<IStorageInfo> {
    static fieldDescriptors: NamedRecordFieldDescriptor<IStorageInfo>[] = [
        ['block_size', 2],
        ['n_blocks', 2]
    ];
};

interface IProgrammingChunkHeaderInfo {
    key_size: number;
    data_size: number;
    flags: number;
};

class ProgrammingChunkHeaderInfo extends NamedDataRecord<IProgrammingChunkHeaderInfo> {
    static fieldDescriptors: NamedRecordFieldDescriptor<IProgrammingChunkHeaderInfo>[] = [
        ['key_size', 2],
        ['data_size', 2],
        ['flags', 4]
    ];
};

function sleep(timeout: number) {
    return new Promise<void>((resolve) => {
        setTimeout(resolve, timeout * 1000);
    });
};

function create_big_endian_Uint32_buffer(...values: number[]) {
    const result = new Uint8Array(values.length * 4);
    const data = new DataView(result.buffer);
    for (let i = 0; i < values.length; i++) {
        data.setUint32(i * 4, values[i]);
    }
    return result;
};

function create_big_endian_Int32_buffer(...values: number[]) {
    const result = new Uint8Array(values.length * 4);
    const data = new DataView(result.buffer);
    for (let i = 0; i < values.length; i++) {
        data.setInt32(i * 4, values[i]);
    }
    return result;
};

interface IFileListEntry {
    name: string;
    flags: number;
    flag_list: string[];
    size: number;
    size_in_blocks: number;
};

export interface ISerialPort {
    flushInput(): void;
    setBreak(value: boolean): void;
};

export interface IDataReader {
    readData(length: number, timeout: number): Promise<Uint8Array>;
};

export interface IDataSender {
    sendData(data: Uint8Array): void;
};

export interface IFileChunkReader {
    read(size_in_bytes: number): Promise<Uint8Array>;
    readonly byteLength: number;
};

export interface IBinaryFileLoader {
    loadBinaryFile(filepath: string): Promise<Uint8Array>;
};

export interface IFsElement {
    name: string;
    isDirectory: boolean;
};

export interface IFsFile {
    readonly flagFailSafe?: boolean;
    readonly flagSecure?: boolean;
    readonly flagPublicWrite?: boolean;
    readonly flagPublicRead?: boolean;
    readonly flagNoSignatureTest?: boolean;
    readonly flagVendor?: boolean;
    readonly flagStatic?: boolean;

    maxsize?: number;
    certificationFileName?: string;
    fileToken?: number;
};

export class FsFile implements IFsElement, IFsFile {
    readonly isDirectory = false;

    readonly flagFailSafe: boolean;
    readonly flagSecure: boolean;
    readonly flagPublicWrite: boolean;
    readonly flagPublicRead: boolean;
    readonly flagSystemFile: boolean;
    readonly flagNoValidCopy: boolean;
    readonly flagSystemFileHostAccess: boolean;
    readonly flagNoSignatureTest: boolean;

    constructor(public name: string, public path: string, public maxsize: number, private flags: number, public blocks = 0, public fsize = 0) {
        this.flagFailSafe = (flags & slconsts.SlFileListProperties.NO_FAILSAFE) === 0;
        this.flagSecure = (flags & slconsts.SlFileListProperties.SECURE) !== 0;
        this.flagPublicWrite = (flags & slconsts.SlFileListProperties.PUBLIC_WRITE) !== 0;
        this.flagPublicRead = (flags & slconsts.SlFileListProperties.PUBLIC_READ) !== 0;
        this.flagSystemFile = (flags & slconsts.SlFileListProperties.SYSTEM_FILE) !== 0;
        this.flagNoValidCopy = (flags & slconsts.SlFileListProperties.NO_VALID_COPY) !== 0;
        this.flagSystemFileHostAccess = (flags & slconsts.SlFileListProperties.SYSTEM_FILE_HOST_ACCESS) !== 0;
        this.flagNoSignatureTest = (flags & slconsts.SlFileListProperties.NOSIGNATURE) !== 0;
    };

    toString() {
        return `FsFile: ${this.name}`;
    };
};

export class FsDir implements IFsElement {
    readonly isDirectory = true;

    children: IFsElement[] = [];
    constructor(public name: string) {
    };

    get_dir(name: string): FsDir | undefined {
        for (let i = 0; i < this.children.length; i++) {
            const folder = this.children[i];
            if (folder.isDirectory && folder.name === name) {
                return folder as FsDir;
            }
        }
    };

    add_dir(folder: FsDir) {
        this.children.push(folder);
    };

    add_file(file: FsFile) {
        this.children.push(file);
    }

    toString() {
        return `FsDir: ${this.name}`;
    };

    sort() {
        this.children.sort((a, b) => {
            if (a.isDirectory === b.isDirectory) {
                return a.name.localeCompare(b.name);
            } else {
                return a.isDirectory ? 1 : -1;
            }
        });
        this.children.forEach((child) => {
            if (child.isDirectory) {
                (child as FsDir).sort();
            }
        });
    };

    find(filepath: string): IFsElement | undefined {
        const parts = filepath.split('/');

        if (filepath.startsWith('/')) {
            parts.shift();  // remove extra leading slash.
        }
        if (filepath.endsWith('/')) {
            parts.pop();    // remove extra trailing slash, if any.
        };

        // eslint-disable-next-line @typescript-eslint/no-this-alias
        let result: IFsElement | undefined = this;
        for (let i = 0; i < parts.length && result; i++) {
            if (result && result instanceof FsDir) {
                result = result.children.find(child => child.name === parts[i]);
            } else {
                result = undefined;
                break;
            }
        }
        return result;
    };
};

export class BootLoader {
    private _last_status?: string;

    constructor(private _reader: IDataReader,
        private _sender: IDataSender,
        private _timeout = 15,
        private _status_func = status_func,
        private _chunksize = slconsts.DEFAULT_CHUNK_SIZE,
        private _power_on_func?: () => Promise<void>,
        private _power_off_func?: () => Promise<void>) {
    };

    private _compute_checksum(...data_segments: (number[] | NamedDataRecord<any> | Uint8Array)[]) {
        let result = 0;
        const computeSum = (result: number, byte: number) => {
            return result + byte;
        };

        for (let i = 0; i < data_segments.length; i++) {
            const data = data_segments[i];
            if (data instanceof NamedDataRecord) {
                result = data.asUint8Array.reduce(computeSum, result);
            } else if (data instanceof Uint8Array) {
                result = data.reduce(computeSum, result);
            } else {
                result = data.reduce(computeSum, result);
            }
        }
        return result & 0xFF;
    };

    private _write_packet(...data_segments: (number[] | NamedDataRecord<any> | Uint8Array)[]) {
        const length = data_segments.reduce((result, segment) => result + segment.length, 0);
        logger.debug(`_write_packet len: ${length}`);

        this._write_header(length, this._compute_checksum(...data_segments));
        for (let i = 0; i < data_segments.length; i++) {
            const data = data_segments[i];
            this._write_data(data);
        }
        return this._expect_ack();
    };

    private async _read_packet(length?: number) {
        const [packet_len, checksum] = await this._read_header();

        if (length !== undefined && length !== packet_len) {
            logger.debug(`Requested length=${length}, actual=$packet_len}`);
            throw new Error(`Error, Mismatch between requested and actual packet length (requested=${length}, actual=${packet_len}`);
        }
        const data = await this._read_data(packet_len);

        const calced_checksum: number = this._compute_checksum(data);

        if (calced_checksum !== checksum) {
            logger.debug(`Calculated: 0x${calced_checksum.toString(16)}.  Received: 0x${checksum.toString(16)}`);
            throw new Error('Checksum error on received packet');
        }
        return data;
    };

    private isEqual(data1: Uint8Array, data2: number[] | Uint8Array) {
        if (data1.length === data2.length) {
            return data1.reduce((result, value, index) => result && value === data2[index], true);
        }
        return false;
    };

    private async _expect_ack() {
        logger.debug('Waiting for ack');

        const data = await this._read_data(slconsts.SL_BOOTLDR_OPCODE_ACK.length);
        logger.debug(`Expecting ack: ${data}`);

        if (!this.isEqual(data, slconsts.SL_BOOTLDR_OPCODE_ACK)) {

            if (this.isEqual(data, slconsts.SL_BOOTLDR_OPCODE_NAK)) {
                const ret = await this._get_operation_status();
                throw new BootLoaderNakError(ret, this._translate_status(ret));
            }
            throw new Error('Timed out waiting for ack');
        }
    };

    private async _try_expect_ack() {
        try {
            await this._expect_ack();
            return true;
        } catch (e) {
            return false;
        }
    };

    _read_data(size: number) {
        return this._reader.readData(size, this._timeout);
    };

    private _write_data(data: number[] | NamedDataRecord<any> | Uint8Array) {
        if (data instanceof NamedDataRecord) {
            data = data.asUint8Array;
        } else if (!(data instanceof Uint8Array)) {
            data = new Uint8Array(data);
        }

        this._sender.sendData(data);
    };

    private _write_ack() {
        this._write_data(slconsts.SL_BOOTLDR_OPCODE_ACK);
    };

    private async _read_header(): Promise<[number, number]> {
        const header = HeaderInfo.create(await this._read_data(HeaderInfo.getSize()));

        return [header.data_length - 2, header.checksum];
    };

    private _write_header(length: number, checksum: number) {
        length += 2;
        const buff = HeaderInfo.create();
        buff.data_length = length;
        buff.checksum = checksum;
        this._write_data(buff);

        logger.debug(`write header: length: ${length} checksum: ${checksum}`);
    };

    private _connected = false;
    private _storage_list?: number;

    async connect_with_reset(port: ISerialPort, timeout: number) {
        const saved_timeout = this._timeout;
        this._timeout = timeout;
        try {
            logger.info('Connecting to device');
            this._connected = false;

            if (this._power_off_func) {
                await this._power_off_func();
            }

            logger.info('Power off');

            await sleep(0.5);

            logger.info('Set break signal');
            port.setBreak(true);

            await sleep(0.1);
            port.flushInput();

            if (this._power_on_func) {
                await this._power_on_func();
            } else {
                logger.info('--- Please restart the device ---');
                this._timeout = 0;
            }

            logger.info('Power on');

            await this._expect_ack();
            this._timeout = timeout;

            logger.info('Clear break signal');
            port.setBreak(false);

            logger.info('Connection succeeded');
            this._connected = true;

            logger.debug('Get storage list');
            try {
                await this._write_packet(slconsts.SL_BOOTLDR_OPCODE_GET_STORAGE_LIST);
                this._storage_list = (await this._read_data(1))[0];
                logger.info('Received storage list');
            } catch (e) {
                this._storage_list = undefined;
                throw e;
            }

            this._connected = true;
        } finally {
            port.setBreak(false);
            this._timeout = saved_timeout;
        }
    };

    async connect_device(port: ISerialPort, ) {
        await this.connect_with_reset(port, 15);
        await this._switch_to_nwp_bootloader(port);
    };

    async connect_device_api(port: ISerialPort, fileLoader: IBinaryFileLoader) {
        await this.connect_device(port);

        const chip_version = await this.get_chip_version();

        const z_device = await this._is_z_device(chip_version);
        const chip_rev = await this._chip_rev(chip_version);
        const g3_dev = await this._is_g3();
        const g3_24only = g3_dev && chip_rev === 0xb1;
        const fs_device = await this._is_fs_device(chip_version);

        if (chip_rev < 0x20 && !z_device) {
            throw new Error('Device not supported');
        }

        if (chip_rev <= 0x20 && z_device && !fs_device) {
            const bootldr_mini = await fileLoader.loadBinaryFile('app/bin/bootldr_mini.bin');
            const bootldr_real = await fileLoader.loadBinaryFile('app/bin/bootldr_real.bin');

            await this._load_mini_and_real_bootloader(port, bootldr_mini, bootldr_real);
        }

        let deviceType: string;
        const cc3200 = await this._is_cc3200(chip_version);
        if (cc3200) {
            if (g3_dev) {
                if (g3_24only) {
                    deviceType = fs_device ? 'CC3230SF' : 'CC3230';
                } else {
                    deviceType = fs_device ? 'CC3235SF' : 'CC3235';
                }
            } else {
                deviceType = fs_device ? 'CC3220SF' : 'CC3220';
            }
        } else {
            if (g3_dev) {
                if (g3_24only) {
                    deviceType = 'CC3130';
                } else {
                    deviceType = 'CC3135';
                }
            } else {
                deviceType = 'CC3120';
            }
        }
        const secure = await this._is_secure(chip_version) || deviceType === 'CC3120' || deviceType === 'CC3135';
        const capacity = await this._get_sflash_capacity();
        const mac_address = await this._get_default_mac_address();

        let dev_not_formatted = false;
        let fs_locked = false;
        let info;

        try {
            info = await this.fs_usage_info();
        } catch (ex) {
            if (ex instanceof BootLoaderNakError) {
                const code = ex.code;
                if (code === (fs_error_codes.FS_NO_DEVICE_IS_LOADED & 0xFF)) {
                    dev_not_formatted = true;
                } else if (code === (fs_error_codes.FS_PROGRAMMING_IN_PROCESS & 0xFF)) {
                } else if (code === (fs_error_codes.FS_ERR_RESET_DURING_PROGRAMMING & 0xFF)) {
                } else if (code === (fs_error_codes.FS_FILE_SYSTEM_IS_LOCKED & 0xFF)) {
                    fs_locked = true;
                } else {
                    throw ex;
                }
            } else {
                throw ex;
            }
        }

        const [a, b, device_id] = await this._get_storage_info('SFLASH');
        const sflash_codes = [
            '0x' + device_id.manufacturer_code.toString(16),
            '0x' + device_id.memory_type.toString(16),
            '0x' + device_id.capacity_code.toString(16)];

        return {
            device: deviceType,
            secure: secure,
            mac_address: mac_address,
            z_device: z_device,
            chip_rev: chip_rev === 0xb1 ? 0x31 : chip_rev,
            capacity: capacity,
            dev_not_formatted: dev_not_formatted,
            fs_locked: fs_locked,
            sflash_codes: sflash_codes,
            ...info
        };
    };

    private _exec_from_ram() {
        return this._write_packet(slconsts.SL_BOOTLDR_OPCODE_EXEC_FROM_RAM);
    };

    disconnect() {
        logger.debug('Disconnecting from device');
        this._connected = false;
        return this._write_packet(slconsts.SL_BOOTLDR_OPCODE_DISCONNECT);
    };

    private async get_version() {
        await this._write_packet(slconsts.SL_BOOTLDR_OPCODE_GET_VERSION_INFO);
        const info = VersionInfo.create(await this._read_packet(VersionInfo.getSize()));
        this._write_ack();
        return info;
    };

    async ping() {
        logger.debug('Pinging device');
        await this._write_packet(slconsts.SL_BOOTLDR_OPCODE_PING);
        logger.debug('Ping successful');
    };

    private async _get_storage_info(storage: keyof IStorages): Promise<[number, number, IDeviceId]> {
        await this._write_packet(slconsts.SL_BOOTLDR_OPCODE_GET_STORAGE_INFO, slconsts.storages[storage]);

        const storage_info_length = StorageInfo.getSize();
        const deviceId_length = DeviceId.getSize();
        const data = await this._read_packet(storage_info_length + deviceId_length);
        this._write_ack();

        const storage_info = StorageInfo.create(data.slice(0, storage_info_length));
        const block_size = storage_info.block_size;
        const n_blocks = storage_info.n_blocks;
        const device_id = DeviceId.create(data.slice(storage_info_length));

        logger.debug(`storage info (block_size, n_blocks): ${block_size}, ${n_blocks}`);

        return [block_size, n_blocks, device_id];
    };

    private async _get_operation_status() {
        logger.debug('Status request');

        await this._write_packet(slconsts.SL_BOOTLDR_OPCODE_GET_LAST_STATUS);
        const data = await this._read_data(4);

        this._write_ack();
        return data[3];
    };

    private async _expect_status_success() {
        let ret = slconsts.SL_BOOTLDR_OPCODE_RET_ACCESS_IN_PROGRESS;

        while (ret === slconsts.SL_BOOTLDR_OPCODE_RET_ACCESS_IN_PROGRESS) {
            ret = await this._get_operation_status();

            if (ret === slconsts.SL_BOOTLDR_OPCODE_RET_SUCCESS) {
                return;
            }
            if (ret === slconsts.COMMAND_RET_FAIL_OTP_ALREADY_LOCKED_ERROR) {
                logger.info('OTP already locked !!!!!');
                return;
            }
        }
        throw new BootLoaderStatusError(ret, this._translate_status(ret));
    };

    private async _write_raw_chunk(storage: keyof IStorages, offset: number, chunk: Uint8Array) {
        if (chunk.length > slconsts.DEFAULT_CHUNK_SIZE - 16) {
            throw new Error('Error, attempt to transfer chunk greater then device\'s buffer');
        }
        const header = create_big_endian_Uint32_buffer(offset, chunk.length);

        await this._write_packet(slconsts.SL_BOOTLDR_OPCODE_RAW_STORAGE_WRITE, slconsts.storages[storage], header, chunk);
        await this._expect_status_success();
    };

    private async _read_raw_chunk(storage: keyof IStorages, offset: number, length: number) {
        const header = create_big_endian_Uint32_buffer(offset, length);

        await this._write_packet(slconsts.SL_BOOTLDR_OPCODE_RAW_STORAGE_READ, slconsts.storages[storage], header);
        const buff = await this._read_packet();
        if (buff.length !== length) {
            throw new Error(`Error, device is sending ${buff.length} bytes, instead of ${length}`);
        }

        this._write_ack();
        await this._expect_status_success();

        return buff;
    };

    private async _erase_blocks(storage: keyof IStorages, offset: number, n_blocks: number) {
        logger.debug(`Erasing blocks: ${storage} ${offset} ${n_blocks}`);
        await this._write_packet(slconsts.SL_BOOTLDR_OPCODE_RAW_STORAGE_ERASE, slconsts.storages[storage], create_big_endian_Uint32_buffer(offset, n_blocks));
        await this._expect_status_success();
    };

    private async _raw_write(storage: keyof IStorages, offset: number, data: Uint8Array, erase_before_write = true, n_blocks = 0) {
        logger.debug('RawStorageWrite');
        if (!slconsts.storages[storage]) {
            logger.error(`Unknown storage ${storage}`);
            return;
        }

        let max_blocks: number;
        let block_size: number;

        if (storage === 'OTP') {
            block_size = 4096;
            max_blocks = 1;
        } else {
            let deviceId: IDeviceId;
            [block_size, max_blocks, deviceId] = await this._get_storage_info(storage);
        }

        if (n_blocks) {
            max_blocks = n_blocks;
        }

        const calced_offset = offset;
        if (calced_offset < 0) {
            throw new Error(`Error, trying to access data located ${calced_offset} bytes before file system's offset`);
        }

        const start_block = Math.floor(calced_offset / block_size);
        if (Math.floor((calced_offset + data.length - 1) / block_size) > max_blocks) {
            throw new Error('Error, size of the data exceeded the Num Of Blocks ');
        }

        const end_block = Math.floor((calced_offset + data.length - 1) / block_size);
        const blocks_to_erase = end_block - start_block + 1;
        if (n_blocks === 0) {
            logger.debug(`BlockSize is ${block_size}, number of blocks is ${max_blocks}`);

            if (erase_before_write && storage !== 'OTP') {
                await this._erase_blocks(storage, start_block, blocks_to_erase);
            }
        }

        let bytes_written = 0;
        const bytes_to_write = data.length;
        while (bytes_written < bytes_to_write) {
            let chunk_size = bytes_to_write - bytes_written;
            if (chunk_size > slconsts.DEFAULT_CHUNK_SIZE - 16) {
                chunk_size = slconsts.DEFAULT_CHUNK_SIZE - 16;
            }

            await this._write_raw_chunk(storage, bytes_written + calced_offset, data.slice(bytes_written, bytes_written + chunk_size));
            bytes_written += chunk_size;

            this._status_func(`Raw Write ${this._last_status || ''}`, bytes_written, bytes_to_write);
        }
    };

    private async _raw_read(storage: keyof IStorages, offset: number, length: number) {
        logger.info('raw read');

        if (!slconsts.storages[storage]) {
            logger.info(`Unknown storage ${storage}`);
            return;
        }

        const [block_size, max_blocks, deviceId] = await this._get_storage_info(storage);

        let bytes_read = 0;
        const data = new Uint8Array(length);

        const calced_offset = offset;
        if (calced_offset < 0) {
            throw new Error(`Error, trying to access data located ${calced_offset} bytes before file system's offset`);
        }

        while (bytes_read < length) {
            let chunk_size = length - bytes_read;
            if (chunk_size > slconsts.DEFAULT_CHUNK_SIZE) {
                chunk_size = slconsts.DEFAULT_CHUNK_SIZE;
            }

            const chunk = await this._read_raw_chunk(storage, bytes_read + calced_offset, chunk_size);
            data.set(chunk, bytes_read);
            bytes_read += chunk_size;

            this._status_func(`Raw Read ${this._last_status || ''}`, bytes_read, length);
        }

        return data;
    };

    private async _load_using_bootloader(port: ISerialPort, bootloader_file: Uint8Array, image_file: Uint8Array) {
        this._last_status = 'Preparing...';

        await this._raw_write('SRAM', 0, bootloader_file);
        await this._expect_status_success();
        await this._exec_from_ram();
        await this._expect_ack();

        this._last_status = 'Preparing...';
        this._raw_write('SRAM', 0, image_file);
        this._expect_status_success();
        this._exec_from_ram();

        port.setBreak(true);
        try {
            await this._expect_ack();
        } finally {
            port.setBreak(false);
        }
    };

    private async _load_mini_bootloader(port: ISerialPort, mini_bootloader_file: Uint8Array, image_file: Uint8Array) {
        await this._load_using_bootloader(port, mini_bootloader_file, image_file);
        await this.get_version();
    };

    private async _load_mini_and_real_bootloader(port: ISerialPort, mini_bootloader_file: Uint8Array, real_bootloader_file: Uint8Array) {
        await this._switch_to_nwp_bootloader(port);
        await this._load_using_bootloader(port, mini_bootloader_file, real_bootloader_file);
    };

    private async _fs_programming_chunk(key: Uint8Array, data: Uint8Array, flags = 0) {
        logger.debug('_fs_programming_chunk');

        const header = ProgrammingChunkHeaderInfo.create();
        header.key_size = key.length;
        header.data_size = data.length;
        header.flags = flags;
        await this._write_packet(slconsts.SL_BOOTLDR_OPCODE_PROGRAMMING, header, key, data);

        const ret = new DataView((await this._read_data(4)).buffer);
        const status = ret.getInt32(0);

        if (status > 0) {
            return [status, 0];
        } else if (status === 0) {
            logger.debug('Finished OK');
            return [status, 0];
        } else {
            return [ret.getInt16(0), ret.getUint16(2)];
        }
    };

    private async fs_programming(use2btldr: boolean, image_file: Uint8Array | IFileChunkReader, key_file?: Uint8Array, flags = 0) {
        const key_size = 16;

        if (key_file && key_file.byteLength !== key_size) {
            throw new Error(`Secret key too short: ${key_file.byteLength} instead of ${key_size}`);
        }
        const secret = key_file || new Uint8Array(0);

        const capacity = use2btldr ? 2 : await this._get_sflash_capacity();

        const saved_timeout = this._timeout;
        if (!capacity) {
            this._timeout = 120 + this._timeout;
        } else {
            this._timeout = Math.floor(slconsts.FLASH_ERASE_TIME_PER_4K_WORST_CASE * ((capacity * (2 ** 20)) / 4096)) + this._timeout;
        }

        try {
            let ret = 0;
            let ex_err = 0;
            let n_bytes = 0;

            if (image_file) {
                const file_size = image_file.byteLength;

                while (ret < file_size) {
                    let chunk_data: Uint8Array;
                    if (image_file instanceof Uint8Array) {
                        chunk_data = image_file.slice(n_bytes, n_bytes + slconsts.DEFAULT_CHUNK_SIZE);
                    } else {
                        chunk_data = await image_file.read(slconsts.DEFAULT_CHUNK_SIZE);
                    }
                    if (!chunk_data) {
                        break;
                    }

                    n_bytes += chunk_data.byteLength;

                    [ret, ex_err] = await this._fs_programming_chunk(secret, chunk_data, flags);

                    if (ret < 0) {
                        break;
                    }

                    if (ret === 0) {
                        ret = file_size;
                    }

                    this._status_func('Image Programming', ret, file_size);
                }
            } else {
                this._status_func('Image Activating', 1, 1);
                [ret, ex_err] = await this._fs_programming_chunk(secret, new Uint8Array(0), 0);
            }

            if (ret < 0) {
                throw new Error(`fs_programming error: ret: ${ret}, ex_err: ${ex_err} - ${slconsts.fs_error_codes[FS_ERROR(ret)]}`);
            }
        } finally {
            this._timeout = saved_timeout;
        }
    };

    private async _is_cc3200(chip_version?: number) {
        if (chip_version === undefined) {
            chip_version = (await this.get_version()).ChipVersion;
        }

        return (chip_version! & 0x10) !== 0;
    };

    private async _is_secure(chip_version?: number) {
        if (chip_version === undefined) {
            chip_version = (await this.get_version()).ChipVersion;
        }

        return (chip_version! & 0x08) !== 0;
    };


    private async _is_24only(chip_version?: number) {
        if (chip_version === undefined) {
            chip_version = (await this.get_version()).ChipVersion;
        }

        return (chip_version! & 0x04) !== 0;
    };

    private async _is_z_device(chip_version?: number) {
        if (chip_version === undefined) {
            chip_version = (await this.get_version()).ChipVersion;
        }

        return ((chip_version! & CHIP_CC3100Z_FUNC_MODE) === CHIP_CC3100Z_FUNC_MODE);
    };

    private async _is_fs_device(chip_version?: number) {
        if (chip_version === undefined) {
            chip_version = (await this.get_version()).ChipVersion;
        }

        const TDFLASH_MASK = 0x03;
        if (0x01 === (chip_version! & TDFLASH_MASK)) {
            return (chip_version! & CHIP_CC3200FS_FUNC_MODE) === CHIP_CC3200FS_FUNC_MODE;
        } else {
            return false;
        }
    };

    private async _is_g3(bl_version?: number) {
        if (bl_version === undefined) {
            bl_version = (await this.get_version()).BootLoaderVersion;
        }

        return ((bl_version! & 0xF000000) >> 20) === 0x40;
    };

    private async _chip_rev(chip_version?: number) {
        if (chip_version === undefined) {
            chip_version = (await this.get_version()).ChipVersion;
        }

        return (chip_version! & 0xFFFF) >> 8;
    };

    async get_chip_version() {
        return (await this.get_version()).ChipVersion;
    };

    private async get_bl_version() {
        return (await this.get_version()).BootLoaderVersion;
    };

    private _switch_uart_to_apps() {
        logger.debug('Switch UART pinmux to APPS');
        const delay = 26666667; // ~ 1 sec delay by the APPS MCU
        return this._write_packet(slconsts.SL_BOOTLDR_OPCODE_SWITCH_2_APPS, create_big_endian_Uint32_buffer(delay));
    };

    async program_flash(filename: string, image_buff: Uint8Array, verify = false) {
        logger.debug('downloading file [NWP_CODE]');
        await this._erase_storage('FLASH');

        logger.info('Writing Flash');
        this._last_status = 'Programming SP';
        await this._raw_write('FLASH', 0, image_buff);

        if (verify) {
            logger.info('Verifying Data...');
            this._last_status = 'Verifying SP';
            const buff = await this._raw_read('FLASH', 0, image_buff.length);
            if (!buff || buff.length !== image_buff.length) {
                throw new Error('Data verification failed');
            }
            for (let i = 0; i < image_buff.length; i++) {
                if (image_buff[i] !== buff[i]) {
                    throw new Error('Data verification failed');
                }
            }
            logger.info('Verification OK');
        }
    };

    private _erase_storage(storage: keyof IStorages, capacity = 256, reserved_blocks_bitmap?: number, format_flags?: number) {
        logger.info(`Erase storage ${storage}`);

        if (format_flags !== undefined) {
            if (format_flags & 0x1) {
                logger.info('Forcing secure FS');
            }
        }

        const fields_to_set = (format_flags === undefined ? 0 : 2) + (reserved_blocks_bitmap === undefined ? 0 : 1);
        const fields_stream = create_big_endian_Uint32_buffer(format_flags || 0, reserved_blocks_bitmap || 0, fields_to_set);

        return this._write_packet(slconsts.SL_BOOTLDR_OPCODE_ERASE, slconsts.storages[storage], fields_stream);
    };

    private async _switch_to_nwp_bootloader(port: ISerialPort) {
        if (await this._is_cc3200()) {
            await this._switch_uart_to_apps();
        }

        try {
            let success = false;
            let i = 0;
            for (; i < 4 && !success; i++) {
                port.setBreak(true);
                sleep(0.100);
                logger.debug('Setting Break Signal');
                if (await this._try_expect_ack()) {
                    success = true;
                }
            }
            if (!success) {
                logger.debug(`--- Connection failed trying to reconnect... (${i + 1} attempt)`);
                throw new Error('Error, Connection failed after 4 attempts');
            }
        } finally {
            port.setBreak(false);
        }

        await this.get_version();
    };

    private async _open_file(file_properties: FileProperties, write_enabled = true) {
        const file_mode = this._generate_access_mode(write_enabled, file_properties);
        const token = create_big_endian_Uint32_buffer(file_mode, file_properties.token || 0);

        const saved_timeout = this._timeout;
        this._timeout = slconsts.FLASH_ERASE_TIME_PER_4K_WORST_CASE * Math.floor(file_properties.size_in_bytes / 4096) + this._timeout;
        try {
            await this._write_packet(slconsts.SL_BOOTLDR_OPCODE_START_DOWNLOAD, token, this._utf8_encoder.encode(file_properties.filename), [0]);

            file_properties.token = new DataView((await this._read_data(4)).buffer).getUint32(0);
            return file_properties.token;
        } finally {
            this._timeout = saved_timeout;
        }
    };

    private async _close_file(file_properties = new FileProperties('   ')) {
        const cert_filename = new Uint8Array(180);
        cert_filename.fill(0);
        cert_filename.set(this._utf8_encoder.encode(file_properties.certificate_filename));

        await this._write_packet(slconsts.SL_BOOTLDR_OPCODE_CLOSE_FILE, cert_filename, this._utf8_encoder.encode(file_properties.signature), [0]);
    };

    private async _get_device_param(param_id: DeviceParams) {
        await this._write_packet(slconsts.SL_BOOTLDR_OPCODE_GET_DEVICE_PARAM, create_big_endian_Uint32_buffer(param_id));

        const data = await this._read_packet();
        this._write_ack();
        return data;
    };

    private async _get_default_mac_address() {
        const data = await this._get_device_param(DeviceParams.DEVICE_PARAM_DEFAULT_MAC_ADDR);
        return Array.from(data).map((value) => ('0' + value.toString(16)).substr(-2)).join(':');
    };

    private async _get_programmable_mac_address() {
        const data = await this._get_device_param(DeviceParams.DEVICE_PARAM_PROGRAMABLE_MAC_ADDR);
        return Array.from(data).map((value) => ('0' + value.toString(16)).substr(-2)).join(':');
    };

    private async _get_file_info(file_properties: FileProperties) {
        const token = create_big_endian_Uint32_buffer(file_properties.token || 0);

        await this._write_packet(slconsts.SL_BOOTLDR_OPCODE_GET_FILE_INFO, token, this._utf8_encoder.encode(file_properties.filename), [0]);

        const fi = FileInfo.create(await this._read_packet(FileInfo.getSize()));
        this._write_ack();
        await this._expect_status_success();

        let flags = 0;
        const old_flags = fi.flags;
        for (let i = 0; i < 16; i++) {
            if ((old_flags >> i) & 0x1) {
                flags |= slconsts.file_info_flags_map[i];
            }
        }

        fi.flags = flags;
        return fi;
    };

    private async _get_chunk(offset: number, length: number) {
        await this._write_packet(slconsts.SL_BOOTLDR_OPCODE_GET_CHUNK, create_big_endian_Uint32_buffer(offset, length));

        const chunk = await this._read_packet();
        this._write_ack();
        await this._expect_status_success();

        return chunk;
    };

    private async _read_file_buffer(file_properties: FileProperties) {
        const file_info = await this._get_file_info(file_properties);

        if (file_info.allocated !== 1) {
            throw new Error(`File not found on file system (${file_properties.filename})`);
        }

        let chunk_size = slconsts.DEFAULT_CHUNK_SIZE;
        let bytes_read = 0;
        const size = file_info.size;
        const data = new Uint8Array(size);

        await this._open_file(file_properties, false);

        while (bytes_read < size) {
            if (size - bytes_read < slconsts.DEFAULT_CHUNK_SIZE) {
                chunk_size = size - bytes_read;
            }

            const chunk_data = await this._get_chunk(bytes_read, chunk_size);
            if (chunk_data.length !== chunk_size) {
                throw new Error(`Error read @ offset ${bytes_read} (requested: ${chunk_size}, read: ${chunk_data})`);
            }
            data.set(chunk_data, bytes_read);
            bytes_read += chunk_size;
            this._status_func('Reading File', bytes_read, size);
        }

        await this._close_file(file_properties);

        return data;
    };

    async dev_mode_download_fs_file(name: string, token = 0) {

        const file_props = new FileProperties(name);
        file_props.token = +token || 0;

        try {
            return await this._read_file_buffer(file_props);
        } catch (ex) {
            if (ex instanceof BootLoaderStatusError || ex instanceof BootLoaderNakError) {
                const info = await this.fs_usage_info();
                throw new BootLoaderFileOperationError(info, ex.message);
            } else {
                throw ex;
            }
        }
    };

    private _generate_access_mode(write_enable: boolean, file_properties: FileProperties) {
        let i = 0;
        let gran = 1;
        for (; i < slconsts.file_size_granularities.length; i++) {
            gran = slconsts.file_size_granularities[i];
            if (gran * 255 >= file_properties.size_in_bytes) {
                break;
            }
        }
        if (i >= slconsts.file_size_granularities.length) {
            throw new Error('Error, size out of range!');
        }

        let n_grans = Math.floor(file_properties.size_in_bytes / gran);
        if (file_properties.size_in_bytes % gran) {
            n_grans += 1;
        }

        logger.info(`Granularity conversion: G(${gran}) * N(${n_grans}) = ${gran * n_grans}`);
        const size_bitmap = (i << 8) | n_grans;

        const access_type = write_enable ? slconsts.FsFileAccessTypes.WRITE_CREATE_IF_NOT_EXIST : slconsts.FsFileAccessTypes.READ;

        logger.debug(`Access type is ${access_type}`);

        let flags = 0;
        if (file_properties.mirror_enable) {
            flags |= slconsts.FsFileAccessFlags.OPEN_FLAG_FAILSAFE;
        }

        if (file_properties.secure) {
            flags |= slconsts.FsFileAccessFlags.OPEN_FLAG_SECURE;
        }

        if (file_properties.vendor) {
            flags |= slconsts.FsFileAccessFlags.OPEN_FLAG_VENDOR;
        }

        if (file_properties.is_static) {
            flags |= slconsts.FsFileAccessFlags.OPEN_FLAG_STATIC;
        }

        if (file_properties.no_signature) {
            flags |= slconsts.FsFileAccessFlags.OPEN_FLAG_NO_SIGNATURE_TEST;
        }

        if (file_properties.public_read) {
            flags |= slconsts.FsFileAccessFlags.OPEN_FLAG_PUBLIC_READ;
        }

        if (file_properties.public_write) {
            flags |= slconsts.FsFileAccessFlags.OPEN_FLAG_PUBLIC_WRITE;
        }

        return (flags << 16) | (access_type << 12) | size_bitmap;
    };

    private _send_chunk(buff: Uint8Array | number[], offset: number) {
        return this._write_packet(slconsts.SL_BOOTLDR_OPCODE_SEND_DATA, create_big_endian_Uint32_buffer(offset), buff);
    };

    private async _write_file_buffer(file_properties: FileProperties, data: Uint8Array | number[]) {
        const length = data.length;
        file_properties.size_in_bytes = Math.max(file_properties.size_in_bytes, length);
        if (length > slconsts.MAX_FILE_SIZE) {
            throw new Error(`Error, size if data (${length}) is too big to write`);
        }

        const token = await this._open_file(file_properties);
        let offset = 0;
        while (offset < length) {
            const buff = data.slice(offset, offset + this._chunksize);
            await this._send_chunk(buff, offset);
            offset += buff.length;
            this._status_func('Writing File', offset, length);
        }

        if (offset !== length) {
            throw new Error(`Error, only ${offset}/${length} bytes were written`);
        }

        await this._close_file(file_properties);
        return token;
    };

    async dev_mode_upload_fs_file(fullPath: string, props: IFsFile, source: Uint8Array | string) {

        const file_props = new FileProperties(
            fullPath,
            props.maxsize,
            props.flagFailSafe,
            props.flagSecure,
            props.flagVendor,
            props.flagStatic,
            props.flagNoSignatureTest,
            props.flagPublicWrite,
            props.flagPublicRead,
            props.certificationFileName
        );

        if (props.flagSecure && !props.flagNoSignatureTest) {
            throw new Error('Invalid Signature');
        }

        let token = props.fileToken;
        if (token) {
            file_props.token = +token;
        }

        if (file_props.size_in_bytes) {
            file_props.size_in_bytes = +file_props.size_in_bytes;
        }

        const buff = source instanceof Uint8Array ? source : this._utf8_encoder.encode(source);

        file_props.size_in_bytes = Math.max(buff.byteLength, file_props.size_in_bytes);

        let info;
        try {
            token = await this._write_file_buffer(file_props, buff);
            info = await this.fs_usage_info();
        } catch (ex) {
            if (ex instanceof BootLoaderStatusError || ex instanceof BootLoaderNakError) {
                info = await this.fs_usage_info();
                throw new BootLoaderFileOperationError(info);
            }
        }

        return { token: token, info: info };
    };

    private async _get_file_list_chunk(index: number, count: number, max_entry_len: number): Promise<[number, number, Uint8Array]> {
        const flags = slconsts.FsListFlags['FS_GET_FILE_ATTRIBUTES'];

        await this._write_packet(slconsts.SL_BOOTLDR_OPCODE_GET_FILE_LIST, create_big_endian_Int32_buffer(index, count), create_big_endian_Uint32_buffer(max_entry_len, flags));

        let data = await this._read_packet();
        const dataView = new DataView(data.buffer);
        this._write_ack();

        index = dataView.getInt32(0);
        count = dataView.getInt32(4);

        data = data.slice(8);

        if (data.length % max_entry_len) {
            throw new Error(`Invalid list size (${data.length})`);
        }

        return [index, count, data];
    };

    private _utf8_encoder = new TextEncoder();
    private _utf8_decoder = new TextDecoder();

    private async _get_file_list() {
        let index = -1;
        let count: number;
        let data: Uint8Array;

        const entry_size = 180 + (4 * 4);
        const max_n_entries = 10;
        const file_entry_size = 3 * 4;
        const files: IFileListEntry[] = [];

        // eslint-disable-next-line no-constant-condition
        while (true) {
            [index, count, data] = await this._get_file_list_chunk(index, max_n_entries, entry_size);
            if (count <= 0) {
                break;
            }

            for (let i = 0; i < data.length; i = i + entry_size) {
                const buff = data.slice(i, i + entry_size);
                const data_view = new DataView(buff.slice(0, file_entry_size).buffer);
                const size = data_view.getUint32(0, true);
                const flags = data_view.getUint32(4, true);
                const size_in_blocks = data_view.getUint32(8, true);
                let j = file_entry_size;
                while (j < buff.length && buff[j] !== 0) {
                    j++;
                }
                const name = this._utf8_decoder.decode(buff.slice(file_entry_size, j));
                const flag_list = slconsts.SlFileListPropertiesFrom_flags(flags);

                files.push({ name: name, flags: flags, flag_list: flag_list, size: size, size_in_blocks });
            }
        }
        return files;
    };

    async get_file_list(): Promise<FsDir> {
        const root = new FsDir('/');

        const fl = await this._get_file_list();

        let curr_dir = root;
        for (let j = 0; j < fl.length; j++) {
            const file = fl[j];
            let flags = file.flags;
            let orig_size = undefined;
            try {
                const fi = await this._get_file_info(new FileProperties(file.name));
                orig_size = fi.size;
                flags |= fi.flags;
            } catch (ex) {
                if (!(ex instanceof BootLoaderStatusError)) {
                    throw ex;
                }
            }

            const parts = file.name.split('/');

            // File paths should not start or end with "/"
            if ((file.name.startsWith('/') && parts.length === 2)) {
                parts.shift();  // remove extra leading slash.
            }
            if (file.name.endsWith('/')) {
                parts.pop();    // remove extra trailing slash.
            };

            const filename = parts.pop();
            curr_dir = root;

            for (let i = 1; i < parts.length; i++) {
                const part = parts[i];
                let dir = curr_dir.get_dir(part);
                if (!dir) {
                    dir = new FsDir(part);
                    curr_dir.add_dir(dir);
                }
                curr_dir = dir;
            }

            curr_dir.add_file(new FsFile(filename || '.', file.name, file.size, flags, file.size_in_blocks, orig_size));
        };

        return root;
    };

    private _rename_file(old_name: string, new_name: string, token = 0) {
        const old_filename = new Uint8Array(256);
        old_filename.fill(0);
        old_filename.set(this._utf8_encoder.encode(old_name));

        const new_filename = new Uint8Array(256);
        new_filename.fill(0);
        new_filename.set(this._utf8_encoder.encode(new_name));

        const token_array = new Uint8Array(4);
        const token_data = new DataView(token_array.buffer);
        token_data.setUint32(0, token);

        return this._write_packet(slconsts.SL_BOOTLDR_OPCODE_RENAME_FILE, token_array, old_filename, new_filename);
    };

    private async _get_sflash_capacity() {
        const [, , device_id] = await this._get_storage_info('SFLASH');

        const manufacturer_code = device_id.manufacturer_code;
        if (!sflash_manufacturers_codes.has(manufacturer_code)) {
            logger.info(`Unknown SFLASH manufacturer ID: ${manufacturer_code}`);
            return;
        }

        const capacity = sflash_capacity_codes.get(device_id.capacity_code);
        if (capacity === undefined) {
            return;
        }

        return capacity / 8;
    };

    private async _delete_file(file_properties: FileProperties) {
        const token = create_big_endian_Uint32_buffer(file_properties.token || 0);

        await this._write_packet(slconsts.SL_BOOTLDR_OPCODE_FILE_ERASE, token, this._utf8_encoder.encode(file_properties.filename), [0]);  // append terminating null charater.
        await this._expect_status_success();
    };

    async dev_mode_delete_fs_file(name: string, token = 0) {

        const file_props = new FileProperties(name);
        file_props.token = +token || 0;

        try {
            await this._delete_file(file_props);
            const info = await this.fs_usage_info();
            return info;
        } catch (ex) {
            if (ex instanceof BootLoaderStatusError || ex instanceof BootLoaderNakError) {
                const info = await this.fs_usage_info();
                throw new BootLoaderFileOperationError(info, ex.message);
            } else {
                throw ex;
            }
        }
    };

    private _translate_status(code: number): string {
        if (code === COMMAND_RET_API_BLOCKED) {
            return 'API Blocked/File System Locked';
        }

        return slconsts.fs_error_codes[FS_ERROR(256 - code)];
    };

    private async _get_fs_usage_info() {
        await this._write_packet(slconsts.SL_BOOTLDR_OPCODE_GET_FS_USAGE);

        const info = FileSysControlDeviceInfo.create(await this._read_packet(FileSysControlDeviceInfo.getSize()));

        this._write_ack();

        return info;
    };

    async fs_usage_info() {
        const info = await this._get_fs_usage_info();
        const dev_mode = await this._is_dev_mode(info);
        const storage_size = Math.floor((info.DeviceBlocksCapacity * info.DeviceBlockSize) / (1 << 10));
        const alerts = await this._get_alerts(info);
        const available_user_files_size = Math.floor((info.NumOfAvailableBlocksForUserFiles * info.DeviceBlockSize) / (1 << 10));

        return { dev_mode: dev_mode, storage_size: storage_size, alerts: alerts, available_user_files_size: available_user_files_size };
    };

    private async _is_dev_mode(info?: IFileSysControlDeviceInfo) {
        if (!info) {
            info = await this._get_fs_usage_info();
        }

        return info.IsDevlopmentFormatType !== 0;  // convert to boolean.
    };

    private async _get_alerts(info?: IFileSysControlDeviceInfo) {
        if (!info) {
            info = await this._get_fs_usage_info();
        }

        return [info.NumOfAlerts, info.NumOfAlertsThreshold];
    };

    private async _load_sec_patch_to_ram_gen3(port: ISerialPort, image: Uint8Array, switch_to_nwp = false) {
        if (switch_to_nwp) {
            await this._switch_to_nwp_bootloader(port);
        }

        const saved_timeout = this._timeout;
        this._timeout = 3;
        try {
            this._last_status = 'Preparing...';
            await this._raw_write('SRAM', 0, image);
            await this._expect_status_success();
            await this._exec_from_ram();
            await this._expect_ack();
            await this.get_version();
        } finally {
            this._timeout = saved_timeout;
        }
    };

    private async _load_sec_patch_to_ram(port: ISerialPort, image: Uint8Array, switch_to_nwp = false) {
        if (switch_to_nwp) {
            await this._switch_to_nwp_bootloader(port);
        }

        const saved_timeout = this._timeout;
        this._timeout = 3;
        try {
            this._last_status = 'Preparing...';
            await this._raw_write('SRAM', 0, image);
            await this._expect_status_success();
            await this._exec_from_ram();
            await this._expect_ack();
            await this.get_version();
        } finally {
            this._timeout = saved_timeout;
        }
    };

    private async _load_patch_to_ram(port: ISerialPort, image: Uint8Array, switch_to_nwp = false) {
        if (switch_to_nwp) {
            await this._switch_to_nwp_bootloader(port);
        }

        const saved_timeout = this._timeout;
        this._timeout = 3;
        try {
            this._last_status = 'Preparing...';
            await this._raw_write('SRAM', 0, image);
            await this._expect_status_success();
            await this._exec_from_ram();
            await this._expect_ack();
            await this.get_version();
        } finally {
            this._timeout = saved_timeout;
        }
    };

    private async _load_otp_file(image: Uint8Array) {
        const saved_timeout = this._timeout;
        this._timeout = 3;
        try {
            this._last_status = 'Preparing...';
            await this._raw_write('OTP', 0, image);
            await this._expect_status_success();
        } finally {
            this._timeout = saved_timeout;
        }
    };

    private _write_patch_to_sflash(image: Uint8Array) {
        const BTL_PTC_BLOCK_OFFSET_BYTE = (33 * 4096) + 8;

        this._last_status = 'Preparing...';
        return this._raw_write('SFLASH', BTL_PTC_BLOCK_OFFSET_BYTE, image);
    };

    private async _restore_to_factory_defaults(defaults_only = false) {
        const FACTORY_RET_TO_IMAGE = 0;
        const FACTORY_RET_TO_DEFAULT = 2;

        let operation = FACTORY_RET_TO_IMAGE;
        if (defaults_only) {
            operation = FACTORY_RET_TO_DEFAULT;
        }

        const capacity = await this._get_sflash_capacity();
        const saved_timeout = this._timeout;
        if (!capacity) {
            this._timeout = 120 + this._timeout;
        } else {
            this._timeout = Math.floor(slconsts.FLASH_ERASE_TIME_PER_4K_WORST_CASE * ((capacity * (2 ** 20)) / 4096)) + this._timeout;
        }

        try {
            logger.info(`Restoring to factory configuration, ${defaults_only ? 'Defaults only' : 'Defaults and Image '} ${operation}`);

            await this._write_packet(slconsts.SL_BOOTLDR_OPCODE_FACTORY_DEFAULTS, create_big_endian_Uint32_buffer(operation));
            const ret = new DataView((await this._read_packet(4)).buffer).getUint32(0);
            if (ret) {
                throw new Error(`Failed with retunr code: ${ret}`);
            }
        } finally {
            this._timeout = saved_timeout;
        }
    };
};

