/**
 *  Copyright (c) 2020, Texas Instruments Incorporated
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:\
 *
 *  *   Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *  notice, this list of conditions and the following disclaimer in the
 *  documentation and/or other materials provided with the distribution.
 *  *   Neither the name of Texas Instruments Incorporated nor the names of
 *  its contributors may be used to endorse or promote products derived
 *  from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 *  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

export enum NamedRecordFieldDataType {
    Int32 = -4,
    Int16 = -2,
    Int8 = -1,
    Uint8 = 1,
    Uint16 = 2,
    Uint32 = 4,
};

export const Int32 = NamedRecordFieldDataType.Int32;
export const Int16 = NamedRecordFieldDataType.Int16;
export const Int8 = NamedRecordFieldDataType.Int8;
export const Uint32 = NamedRecordFieldDataType.Uint32;
export const Uint16 = NamedRecordFieldDataType.Uint16;
export const Uint8 = NamedRecordFieldDataType.Uint8;

export class Utf8cString {
    constructor(public size: number) {
        if (size <= 1) {
            throw new Error('Programmer Error: Utf8cString must have a size greater than 1');
        }
    }
};

export class Union<T> {
    unionDescriptors: NamedRecordFieldDescriptor<T>[][];
    size: number;
    constructor(...fieldDescriptors: NamedRecordFieldDescriptor<T>[][]) {
        this.unionDescriptors = fieldDescriptors;
        // eslint-disable-next-line @typescript-eslint/no-use-before-define
        this.size = this.unionDescriptors.reduce((result, descriptors: NamedRecordFieldDescriptor<T>[]) => Math.max(result, sizeOf(descriptors)), 0);
    }
};

export type NamedRecordFieldDescriptor<T> = [keyof T | 'reserved' | 'padding', NamedRecordFieldDataType | Utf8cString, number?] | Union<T>;

type Factory<T> = {
    new(fieldDescriptors: NamedRecordFieldDescriptor<T>[], buffer?: Uint8Array | number[], littleEndian?: boolean): NamedDataRecord<T>;
    fieldDescriptors: NamedRecordFieldDescriptor<T>[];
};

function sizeOf<T>(fieldDescriptors: NamedRecordFieldDescriptor<T>[]): number {
    return fieldDescriptors.reduce((result, fieldDescriptor) => {
        if (fieldDescriptor instanceof Union) {
            return result + fieldDescriptor.size;
        }
        const [key, type, count = 1] = fieldDescriptor;
        const size: number = type instanceof Utf8cString ? (type as Utf8cString).size : Math.abs(type);
        return result + size * count;
    }, 0);
};

export class NamedDataRecord<T> {
    private dataView: DataView;
    private rawData: Uint8Array;
    protected littleEndian = false;
    constructor(fieldDescriptors: NamedRecordFieldDescriptor<T>[], buffer?: Uint8Array | number[]) {
        const offset = sizeOf(fieldDescriptors);
        buffer = buffer || new Uint8Array(offset);
        this.rawData = buffer instanceof Uint8Array ? buffer : new Uint8Array(buffer);

        if (offset !== this.rawData.length) {
            throw new Error('Programmer Error, the size of the buffer passed in the construtor of NamedDataRecord does not match the NamedRecordField descriptors.');
        };

        this.dataView = new DataView(this.rawData.buffer);

        this.defineFields(offset, fieldDescriptors);
    };

    private defineFields(offset: number, fieldDescriptors: NamedRecordFieldDescriptor<T>[]) {

        for (let i = fieldDescriptors.length; i-- > 0;) {
            const descriptor = fieldDescriptors[i];
            if (descriptor instanceof Union) {
                const union = descriptor as Union<T>;
                offset = offset - union.size;
                union.unionDescriptors.forEach((unionDescriptor) => {
                    this.defineFields(offset + sizeOf(unionDescriptor), unionDescriptor);
                });
            } else {
                const [key, type, count = 1] = descriptor;

                const size: number = type instanceof Utf8cString ? (type as Utf8cString).size : Math.abs(type);
                offset = offset - size * count;

                if (key !== 'reserved' && key !== 'padding') {
                    if (key in this) {
                        throw new Error(`Programmer Error, the field name ${key} has already been used.  Either it is a duplicate, or it is a reserved keyword`);
                    }

                    if (type instanceof Utf8cString) {
                        this.defineStringField(key, type, offset, count);
                    } else if (count && count > 1) {
                        this.defineArrayField(key, type, offset, count);
                    } else {
                        this.defineField(key, type, offset);
                    }
                }
            }
        }
    };

    private defineField(key: keyof T, size: NamedRecordFieldDataType, offset: number) {
        switch (size) {
            case 1:
                Object.defineProperty(this, key, {
                    get: () => {
                        return this.dataView.getUint8(offset);
                    },
                    set: (value: number) => {
                        this.dataView.setUint8(offset, value);
                    }
                });
                break;
            case 2:
                Object.defineProperty(this, key, {
                    get: () => {
                        return this.dataView.getUint16(offset, this.littleEndian);
                    },
                    set: (value: number) => {
                        this.dataView.setUint16(offset, value, this.littleEndian);
                    }
                });
                break;
            case 4:
                Object.defineProperty(this, key, {
                    get: () => {
                        return this.dataView.getUint32(offset, this.littleEndian);
                    },
                    set: (value: number) => {
                        this.dataView.setUint32(offset, value, this.littleEndian);
                    }
                });
                break;
            case -1:
                Object.defineProperty(this, key, {
                    get: () => {
                        return this.dataView.getInt8(offset);
                    },
                    set: (value: number) => {
                        this.dataView.setInt8(offset, value);
                    }
                });
                break;
            case -2:
                Object.defineProperty(this, key, {
                    get: () => {
                        return this.dataView.getInt16(offset, this.littleEndian);
                    },
                    set: (value: number) => {
                        this.dataView.setInt16(offset, value, this.littleEndian);
                    }
                });
                break;
            case -4:
                Object.defineProperty(this, key, {
                    get: () => {
                        return this.dataView.getInt32(offset, this.littleEndian);
                    },
                    set: (value: number) => {
                        this.dataView.setInt32(offset, value, this.littleEndian);
                    }
                });
                break;
            default:
                throw new Error(`Programmer Error, the size for field ${key} is not supportted.`);
        }
    };

    private assertArrayLength(key: keyof T, expected: number, actual: number) {
        if (expected !== actual) {
            throw `Programmer Error, cannot set array field ${key} of size ${expected} because the array size does not match.`;
        }
    };

    private defineArrayField(key: keyof T, type: NamedRecordFieldDataType, offset: number, count: number) {
        const size = Math.abs(type);
        switch (type) {
            case 1:
                Object.defineProperty(this, key, {
                    get: () => {
                        return new Uint8Array(this.rawData.slice(offset, offset + count * size));
                    },
                    set: (value: number[] | Uint8Array) => {
                        this.assertArrayLength(key, count, value.length);
                        this.rawData.set(value, offset);
                    }
                });
                break;
            case 2:
                Object.defineProperty(this, key, {
                    get: () => {
                        if (this.littleEndian) {
                            const data = this.rawData.slice(offset, offset + count * size);
                            return new Uint16Array(data.buffer);
                        }
                        const result = new Uint16Array(count * size);
                        for (let i = 0; i < count; i++) {
                            result.set([this.dataView.getUint16(offset + i * size)], i);
                        }
                        return result;
                    },
                    set: (value: number[] | Uint16Array) => {
                        this.assertArrayLength(key, count, value.length);
                        if (this.littleEndian) {
                            this.rawData.set(value, offset);
                        } else {
                            for (let i = 0; i < count; i++) {
                                this.dataView.setUint16(offset + i * size, value[i]);
                            }
                        }
                    }
                });
                break;
            case 4:
                Object.defineProperty(this, key, {
                    get: () => {
                        if (this.littleEndian) {
                            const data = this.rawData.slice(offset, offset + count * size);
                            return new Uint32Array(data.buffer);
                        }
                        const result = new Uint32Array(count * size);
                        for (let i = 0; i < count; i++) {
                            result.set([this.dataView.getUint32(offset + i * size)], i);
                        }
                        return result;
                    },
                    set: (value: number[] | Uint32Array) => {
                        this.assertArrayLength(key, count, value.length);
                        if (this.littleEndian) {
                            this.rawData.set(value, offset);
                        } else {
                            for (let i = 0; i < count; i++) {
                                this.dataView.setUint32(offset + i * size, value[i]);
                            }
                        }
                    }
                });
                break;
            case -1:
                Object.defineProperty(this, key, {
                    get: () => {
                        const data = this.rawData.slice(offset, offset + count * size);
                        return new Int8Array(data.buffer);
                    },
                    set: (value: number[] | Int8Array) => {
                        this.assertArrayLength(key, count, value.length);
                        this.rawData.set(value, offset);
                    }
                });
                break;
            case -2:
                Object.defineProperty(this, key, {
                    get: () => {
                        if (this.littleEndian) {
                            const data = this.rawData.slice(offset, offset + count * size);
                            return new Int16Array(data.buffer);
                        }
                        const result = new Int16Array(count * size);
                        for (let i = 0; i < count; i++) {
                            result.set([this.dataView.getInt16(offset + i * size)], i);
                        }
                        return result;
                    },
                    set: (value: number[] | Int16Array) => {
                        this.assertArrayLength(key, count, value.length);
                        if (this.littleEndian) {
                            this.rawData.set(value, offset);
                        } else {
                            for (let i = 0; i < count; i++) {
                                this.dataView.setInt16(offset + i * size, value[i]);
                            }
                        }
                    }
                });
                break;
            case -4:
                Object.defineProperty(this, key, {
                    get: () => {
                        if (this.littleEndian) {
                            const data = this.rawData.slice(offset, offset + count * size);
                            return new Int32Array(data.buffer);
                        }
                        const result = new Int32Array(count * size);
                        for (let i = 0; i < count; i++) {
                            result.set([this.dataView.getInt32(offset + i * size)], i);
                        }
                        return result;
                    },
                    set: (value: number[] | Int32Array) => {
                        this.assertArrayLength(key, count, value.length);
                        if (this.littleEndian) {
                            return this.rawData.set(value, offset);
                        } else {
                            for (let i = 0; i < count; i++) {
                                this.dataView.setInt32(offset + i * size, value[i]);
                            }
                        }
                    }
                });
                break;
        }
    };

    private utf8Decoder?: TextDecoder;
    private utf8Encoder?: TextEncoder;

    private getStringFieldValue(offset: number, size: number) {
        let len = 0;
        while (len < size-1 && this.rawData[offset + len] !== 0) {
            len++;
        }
        if (len > 0) {
            return this.utf8Decoder!.decode(this.rawData.slice(offset, offset + len));
        }
        return '';
    };

    private setStringFieldValue(value: string, offset: number, size: number, key: keyof T) {
        const utf8Value = this.utf8Encoder!.encode(value);
        if (!utf8Value) {
            throw new Error(`Programmer Error: Cannot convert string ${value} to a utf-8 string`);
        }
        if (utf8Value.length >= size) {
            throw new Error(`Programmer Error: Cannot set string field ${key} because of buffer overrun`);
        }
        this.rawData.set(utf8Value, offset);
        this.rawData.fill(0, offset + utf8Value.length, offset + size);
    };

    private defineStringField(key: keyof T, type: Utf8cString, offset: number, count: number = 1) {
        if (!this.utf8Decoder) {
            this.utf8Decoder = new TextDecoder();
        }
        if (!this.utf8Encoder) {
            this.utf8Encoder = new TextEncoder();
        }
        if (count > 1) {
            Object.defineProperty(this, key, {
                get: () => {
                    const result = new Array<string>(count);
                    for (let i = 0; i < count; i++) {
                        result[i] = this.getStringFieldValue(offset + i * type.size, type.size);
                    }
                    return result;
                },
                set: (value: string[]) => {
                    this.assertArrayLength(key, count, value.length);
                    for (let i = 0; i < count; i++) {
                        this.setStringFieldValue(value[i], offset + i * type.size, type.size, key);
                    }
                }
            });
        } else {
            Object.defineProperty(this, key, {
                get: () => this.getStringFieldValue(offset, type.size),
                set: (value: string) => {
                    this.setStringFieldValue(value, offset, type.size, key);
                }
            });
        }
    };

    get asUint8Array(): Uint8Array {
        return this.rawData;
    };

    get length() {
        return this.dataView.buffer.byteLength;
    };

    static getSize<T>(this: { fieldDescriptors: NamedRecordFieldDescriptor<T>[] }) {
        return sizeOf(this.fieldDescriptors);
    };

    static create<T>(this: Factory<T>, buffer?: Uint8Array | number[]) {
        return new this(this.fieldDescriptors, buffer) as unknown as (T & NamedDataRecord<T>);
    };
};
