/**
 *  Copyright (c) 2020, Texas Instruments Incorporated
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:\
 *
 *  *   Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *  notice, this list of conditions and the following disclaimer in the
 *  documentation and/or other materials provided with the distribution.
 *  *   Neither the name of Texas Instruments Incorporated nor the names of
 *  its contributors may be used to endorse or promote products derived
 *  from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 *  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
import { BootLoader, IDataReader, ISerialPort, IDataSender, setLogger, IFsFile } from './BootLoader';
import { xds110reset } from './xds110reset';

declare namespace gc {
    const services: { [index: string]: unknown };
    namespace fileCache {
        function readBinaryFile(url: string): any;
    }
    namespace databind {
        function registerCustomCodec(name: string, factory: any): void;
        let registry: any;
    }
    namespace console {
        function info(module: string, message: any): void;
        function error(module: string, message: any): void;
        function debug(module: string, message: any): void;
    }
}

declare const Q: any;

document.addEventListener('gc-databind-ready', function () {

    setLogger('debug', gc.console.debug.bind(gc.console, 'ImageCreatorCodec'));
    setLogger('info', gc.console.info.bind(gc.console, 'ImageCreatorCodec'));
    setLogger('error', gc.console.error.bind(gc.console, 'ImageCreatorCodec'));

    function power_on() {
        return xds110reset(false, gc.console.error.bind(gc.console, 'ImageCreatorCodec'));
    };

    function power_off() {
        return xds110reset(true, gc.console.error.bind(gc.console, 'ImageCreatorCodec'));
    };

    // TODO: remove this when TiFiles stops supporting Blob.
    function convertBlob2ArrayBuffer(promise: any): Promise<Uint8Array> {
        return new Promise((resolve, reject) => {
            promise.then((data: Blob) => {
                const reader = new FileReader();
                reader.onloadend = () => {
                    resolve(new Uint8Array(reader.result as ArrayBuffer));
                };
                reader.readAsArrayBuffer(data);

            }).fail(reject);
        });
    };

    const binaryFileLoader = new (class {
        async loadBinaryFile(filepath: string) {
            return new Uint8Array(await convertBlob2ArrayBuffer(gc.fileCache.readBinaryFile(filepath)));
        }
    })();

    /**
     * This code implements the IPacketCodec interface.
     *
     * @constructor
     * @implements gc.databind.IPacketCodec
     */
    class CustomCodec implements IDataReader, ISerialPort, IDataSender {
        private requiredDataLength = 0;
        private abortAllOperations = false;
        private sequencer = Promise.resolve(); // sequencer for queuing operations to perform one after the other.
        private debouncePromises = new Map();
        private boot_loader: BootLoader;
        private targetDecoder: { decode: (data: object) => boolean } = {
            decode: () => {
                return false;
            }
        };
        private partialPacket?: Uint8Array;
        private deferedReadPromise: any;
        private is_connected = false;
        private serialIO: any;
        readonly busy = false;
        private busyCount = 0;

        constructor() {
            this.boot_loader = new BootLoader(this, this, 15, () => { }, undefined, power_on, power_off);
        };

        readFileContents = this.sequenceOperation(async (filepath) => {
            return await this.boot_loader.dev_mode_download_fs_file(filepath);
        });

        writeFileContents = this.sequenceOperation(async (filepath, contents: ArrayBuffer, options?: IFsFile) => {
            await this.boot_loader.dev_mode_upload_fs_file(filepath, options || {}, new Uint8Array(contents));
        });

        deleteFileContents = this.sequenceOperation(async (filepath) => {
            await this.boot_loader.dev_mode_delete_fs_file(filepath);
        });

        /**
         * Encodes data into a packet for sending to the target device.
         *
         * @param target - function to call to pass on the encoded data towards the target
         * @param jsonObj - The data object from the GUI that is to be encoded into a packet for sending.
         */
        encode(target: (data: number[]) => void, data: number[]) {
            target(data);
        };

        /**
         * Decodes packets from the target device into data objects.  One object for each packet of data.
         * Unless you have chained your custom codec with the CR codec, the packet data is not framed,
         * so this method is responsible for decoding partial packets as well as multiple packets.  Partial packets
         * should not be returned, but deferred until the next decode method call with more raw data to process.  As a
         * result, this method may return 0 or more packets of data on each invocation.
         *
         * @param target -  function to call to pass on a JSON object containing the decoded values towards the GUI
         * @param data  - The raw data received that needs to be decoded from packets into objects.
         * @return - true if connected (i.e. valid data received), false if not connected
         */
        decode(target: (data: any) => boolean, rx_data: number[]) {
            this.targetDecoder.decode = target;

            // concatentate new data with saved partial data
            if (this.partialPacket) {
                this.partialPacket = new Uint8Array([...this.partialPacket, ...rx_data]);
            } else {
                this.partialPacket = new Uint8Array(rx_data);
            }
            if (this.partialPacket.length >= this.requiredDataLength && this.deferedReadPromise) {
                this.deferedReadPromise.resolve();
            }
            return true;
        };

        readData(size: number, timeout: number) {
            return new Promise<Uint8Array>((resolve, reject) => {
                if (!this.partialPacket || this.partialPacket.length < size) {
                    this.requiredDataLength = size;
                    this.deferedReadPromise = Q.defer();
                    let promise = this.deferedReadPromise.promise;
                    if (timeout > 0) {
                        promise = promise.timeout(timeout * 1000, 'Timeout on reading data from serial port');
                    }
                    promise.then(() => {
                        return this.readData(size, timeout);
                    }).finally(() => {
                        this.requiredDataLength = 0;
                        this.deferedReadPromise = undefined;
                    }).then(resolve).fail(reject);
                } else {
                    const result = this.partialPacket.slice(0, size);
                    this.partialPacket = this.partialPacket.slice(size);
                    resolve(result);
                }
            });
        };

        encoder: any;

        sendData(data: Uint8Array) {
            this.encoder([...data]);
        };

        connect() {
            return new Q.Promise((resolve: (result?: any) => void, reject: (reason: any) => void) => {
                this.abort().then(async () => {
                    try {
                        this.serialIO = gc.services['ti-service-serialio'];

                        const usbTransport: any = document.querySelector('ti-transport-usb');

                        usbTransport.addConsoleProgress('waiting for reset...');
                        const data = await this.boot_loader.connect_device_api(this, binaryFileLoader);

                        this.is_connected = true;
                        if (this.targetDecoder) {
                            this.targetDecoder.decode(data);
                        }
                        resolve();
                        this.refresh();
                    } catch (e) {
                        reject(e.toString());
                    }
                }).catch(reject);
            });
        };

        refresh: () => Promise<void> = this.debounceOperation<void>('refresh', this.sequenceOperation(async () => {
            try {
                const fl = await this.boot_loader.get_file_list();
                const info = await this.boot_loader.fs_usage_info();

                fl.sort();
                this.targetDecoder.decode({ userFiles: fl, ...info });
            } catch (err) {
                const usbTransport: any = document.querySelector('ti-transport-usb');
                const errMsg = err.message || err.toString();
                usbTransport.addConsoleError('Problem reading user files', errMsg, errMsg);
                throw err;
            }
        }));

        private abort() {
            // make sure all queued operations, that are waiting to go, are aborted too
            this.abortAllOperations = true;
            this.sequencer = this.sequencer.finally(() => {
                this.abortAllOperations = false;
            });
            // reject the current operation
            if (this.deferedReadPromise) {
                this.deferedReadPromise.reject('This operation was aborted by the user.');
                this.deferedReadPromise = undefined;
            }
            this.flushInput();
            return this.sequencer;
        };

        setBusy() {
            if (this.busyCount === 0) {
                gc.databind.registry.getBinding('prop.isCodecBusy').setValue(true);
            }
            this.busyCount++;
        };

        clearBusy() {
            this.busyCount--;
            if (this.busyCount === 0) {
                gc.databind.registry.getBinding('prop.isCodecBusy').setValue(false);
            }
        }

        private sequenceOperation<T>(operation: (...params: any) => Promise<T>): () => Promise<T> {
            return (...params: any) => {
                this.setBusy();
                return new Promise<T>( (resolve, reject) => {
                    this.sequencer = this.sequencer.finally(() => {
                        if (this.abortAllOperations) {
                            reject(new Error('This operation was aborted by the user.'));
                        } else {
                            return operation(...params).then(resolve).catch(reject);
                        }
                    });
                }).finally(() => this.clearBusy());
            };
        };

        private debounceOperation<T>(name: string, operation: (...params: any) => Promise<T>): () => Promise<T> {
            return (...params: any) => {
                let result = this.debouncePromises.get(name);
                if (!result) {
                    result = operation(params).finally(()=> {
                        this.debouncePromises.set(name, undefined);
                    });
                    this.debouncePromises.set(name, result);
                }
                return result;
            };
        };

        setBreak(value: boolean) {
            if (this.serialIO) {
                this.serialIO.setSerialSignals({ brk: value });

                if (value === false && !this.is_connected) {
                    const usbTransport: any = document.querySelector('ti-transport-usb');
                    usbTransport.addConsoleProgress('waiting for device status...');
                }
            }
        };

        flushInput() {
            this.requiredDataLength = 0;
            this.partialPacket = undefined;
        };

        disconnect() {
            this.abort();
            return new Q.Promise((resolve: (result?: any) => void) => {
                if (this.is_connected) {
                    this.is_connected = false;
                    this.boot_loader.disconnect().then(resolve);
                }
            });
        };

        convertDataToText(data: Uint8Array): { text: string; encoding: string } {
            // convert UTF-8, UTF-16 little endian, and UTF-16 big endian to text string, and presere the original encoding if possible.
            let text;
            let encoding: string;
            if (data.length % 1 === 0 && data.length > 2 && data[0] === 0xFF && data[1] === 0xFE) {
                // UTF-16 little endian
                encoding = 'UTF-16 LE';
                text = String.fromCharCode(...(new Uint16Array(data.buffer)));
            } else if (data.length % 1 === 0 && data.length > 2 && data[0] === 0xFE && data[1] === 0xFF ) {
                // UTF-16 big endian
                encoding = 'UTF-16 BE';
                const len = data.length/2;
                const buffer = new Array(data.length/2);
                const dataView = new DataView(data.buffer);
                for (let i = 0; i < len; i++) {
                    buffer[i] = dataView.getUint16(i*2);
                }
                text = String.fromCharCode(...buffer);
            } else {
                // 'UTF-8'
                encoding = 'UTF-8';
                const decoder = new TextDecoder();
                text = decoder.decode(data.buffer);
            }
            return { text: text, encoding: encoding };
        };

        convertTextToData(contents: string, encoding: string): ArrayBuffer {

            let buffer: ArrayBuffer;

            // convert UTF-8, UTF-16 little endian, or UTF-16 big endian from text string, based on encoding provided
            if (encoding.startsWith('UTF-16')) {
                const littleEndian = !encoding.endsWith('BE');
                buffer = new ArrayBuffer(contents.length * 2);
                const dataView = new DataView(buffer);
                for (let i = 0; i < contents.length; i++) {
                    dataView.setUint16(i, contents.charCodeAt(i), littleEndian);
                }
            } else {
                const encoder = new TextEncoder();
                buffer = encoder.encode(contents);
            }
            return buffer;
        };

    };

    gc.databind.registerCustomCodec('custom', CustomCodec);
});
